/// <reference types="node" />
import tls from "tls";
import { Logger } from "./logger";
import { Statistics } from "./statistics";
import { ClusterMessage } from "./httpReverseProxy";
/**
 * Options for the certificate manager.
 *
 * certificateStoreRoot: The absolute or relative path to the
 * base directory for the certificates in the file system.
 *
 * This directory must have read, write, delete privedges
 */
export interface CertificateOptions {
    certificateStoreRoot: string;
    log?: Logger;
    stats?: Statistics;
}
/**
 * Additional information about the certificate to allow management of expirations
 */
export interface CertificateInformation {
    expiresOn: Date;
    commonName: string;
}
/**
 * The message format for messaging through the
 * master to other workers in a cluster
 */
export interface CertificateMessage extends ClusterMessage {
    hostname: string;
    key: string;
    certificate: string;
    ca?: string;
}
/**
 * The certificate and addiional data stored in the certificate table
 */
interface CertificateData {
    secureContext: tls.SecureContext;
    certificateInformation?: CertificateInformation;
}
/**
 * the Certificate table. Properties are the hostname,
 * value is the certificate and additional information
 */
interface ActiveCertificates {
    [hostname: string]: CertificateData;
}
/**
 * The class implementing the certificate store
 */
export declare class Certificates {
    private certificateStoreRoot;
    protected certificates: ActiveCertificates;
    protected log: Logger;
    protected stats: Statistics;
    constructor(options: CertificateOptions);
    /**
     * Return all of the active certificates
     */
    getActiveCertificates: () => ActiveCertificates;
    /**
     * Get the certificate for a given hostname
     */
    getCertificate: (hostname: string) => tls.SecureContext;
    /**
     * Get the additional certificate information for a given hostname
     */
    getCertificateInformation: (hostname: string) => CertificateInformation;
    /**
     * Add a certificate to the table if one does not exist
     */
    addCertificate: (hostname: string, secureContext: tls.SecureContext) => boolean;
    /**
     * Update a certificate in the table if it exists
     */
    updateCertificate: (hostname: string, secureContext: tls.SecureContext) => boolean;
    /**
     * Remove a certificate from the table
     */
    removeCertificate: (hostname: string) => boolean;
    /**
     * Load a certificate from a key, cert, and ca strings. This is the format of
     * the certificates on the filesystem and generated from the LetsEncrypt system
     */
    loadCertificate: (hostname: string, key: string | string[], certificate: string | string[], ca?: string | string[]) => boolean;
    /**
     * Load a certificate from a set of PEM encoded files (key, certificate, ca?)
     *
     * Files are absolute locations, not from the certificateStoreRoot
     */
    loadCertificateFromFiles: (hostname: string, keyFilename: string, certificateFileName: string | string[], caFilename?: string | string[]) => boolean;
    /**
     * Load a certificate relative to the certificateStoreRoot based on the hostname
     */
    loadCertificateFromStore: (hostname: string) => boolean;
    /**
     * Save a certificate relative to the certificateStoreRoot based on the hostname
     */
    saveCertificateToStore: (hostname: string, key: string, certificate: string, ca?: string) => boolean;
    /**
     * In a cluster environment propagate the certificate
     * to the other worker processes via the master process
     */
    propagateNewCertificate: (hostname: string, key: string, certificate: string, ca?: string) => void;
    /**
     * In a cluster environment the master will call this method to propagate the certificate
     */
    processMessage: (message: CertificateMessage) => void;
    /**
     * Read a (set) of PEM file(s) and flatten and unbundle them if they are a bundle
     */
    getCertificateData: (pathName: string | string[], unbundle: boolean) => string | string[];
    /**
     * Unbundle a PEM string
     */
    private unbundleCertificate;
    /**
     * File system helper functions
     *
     * All operations are sync due to time constraints at startup
     */
    private exists;
    private mkDir;
    private writeFile;
    private readFile;
}
export {};
