"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const tls_1 = __importDefault(require("tls"));
const cluster_1 = __importDefault(require("cluster"));
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const node_forge_1 = __importDefault(require("node-forge"));
/**
 * The class implementing the certificate store
 */
class Certificates {
    constructor(options) {
        this.certificates = {};
        /**
         * Return all of the active certificates
         */
        this.getActiveCertificates = () => {
            return this.certificates;
        };
        /**
         * Get the certificate for a given hostname
         */
        this.getCertificate = (hostname) => {
            this.stats && this.stats.updateCount('CertificatesRequested', 1);
            return this.certificates[hostname] && this.certificates[hostname].secureContext;
        };
        /**
         * Get the additional certificate information for a given hostname
         */
        this.getCertificateInformation = (hostname) => {
            return this.certificates[hostname] && this.certificates[hostname].certificateInformation;
        };
        /**
         * Add a certificate to the table if one does not exist
         */
        this.addCertificate = (hostname, secureContext) => {
            if (!this.certificates[hostname]) {
                this.stats && this.stats.updateCount('CertificatesAdded', 1);
                this.certificates[hostname] = { secureContext };
                return true;
            }
            return false;
        };
        /**
         * Update a certificate in the table if it exists
         */
        this.updateCertificate = (hostname, secureContext) => {
            if (this.certificates[hostname]) {
                this.stats && this.stats.updateCount('CertificatesUpdated', 1);
                this.certificates[hostname] = { secureContext };
                return true;
            }
            return false;
        };
        /**
         * Remove a certificate from the table
         */
        this.removeCertificate = (hostname) => {
            if (!this.certificates[hostname]) {
                return false;
            }
            this.stats && this.stats.updateCount('CertificatesRemoved', 1);
            delete this.certificates[hostname];
            return true;
        };
        /**
         * Load a certificate from a key, cert, and ca strings. This is the format of
         * the certificates on the filesystem and generated from the LetsEncrypt system
         */
        this.loadCertificate = (hostname, key, certificate, ca) => {
            if (this.certificates[hostname] || !key || !certificate) {
                return false;
            }
            const details = {
                key: key,
                cert: certificate,
            };
            if (ca) {
                details.ca = ca;
            }
            /**
             * Create the secure context used by the https SNI interface.
             */
            const context = tls_1.default.createSecureContext(details).context;
            /**
             * use forge to extract the data from the PEM string
             */
            const cert = node_forge_1.default.pki.certificateFromPem(Array.isArray(certificate) ? certificate[0] : certificate);
            this.stats && this.stats.updateCount('CertificatesLoaded', 1);
            this.certificates[hostname] = {
                secureContext: context,
                certificateInformation: {
                    expiresOn: cert.validity.notAfter,
                    commonName: cert.subject.attributes.reduce((current, attribute) => attribute.name === 'commonName' ? attribute.value : current, null)
                }
            };
            return true;
        };
        /**
         * Load a certificate from a set of PEM encoded files (key, certificate, ca?)
         *
         * Files are absolute locations, not from the certificateStoreRoot
         */
        this.loadCertificateFromFiles = (hostname, keyFilename, certificateFileName, caFilename) => {
            const key = this.getCertificateData(keyFilename, false);
            const certificate = this.getCertificateData(certificateFileName, false);
            const ca = caFilename && this.getCertificateData(caFilename, true);
            return this.loadCertificate(hostname, key, certificate, ca);
        };
        /**
         * Load a certificate relative to the certificateStoreRoot based on the hostname
         */
        this.loadCertificateFromStore = (hostname /*, loadCertificateData?: boolean*/) => {
            const pathName = hostname.replace(/\./g, '_').replace(/\*/g, '');
            const storePath = path_1.default.join(`${this.certificateStoreRoot}`, `${pathName}`);
            const key = this.getCertificateData(path_1.default.join(`${storePath}`, `${pathName}-key.pem`), false);
            const certificate = this.getCertificateData(path_1.default.join(`${storePath}`, `${pathName}-crt.pem`), false);
            const ca = this.getCertificateData(path_1.default.join(`${storePath}`, `${pathName}-ca.pem`), true);
            return this.loadCertificate(hostname, key, certificate, ca); //, loadCertificateData)
        };
        /**
         * Save a certificate relative to the certificateStoreRoot based on the hostname
         */
        this.saveCertificateToStore = (hostname, key, certificate, ca) => {
            delete this.certificates[hostname];
            this.loadCertificate(hostname, key, certificate, ca);
            const pathName = hostname.replace(/\./g, '_');
            const storePath = path_1.default.join(`${this.certificateStoreRoot}`, `${pathName}`);
            this.stats && this.stats.updateCount('CertificatesSavedToStore', 1);
            return this.mkDir(storePath, { recursive: true }) &&
                this.writeFile(path_1.default.join(`${storePath}`, `${pathName}-key.pem`), key) &&
                this.writeFile(path_1.default.join(`${storePath}`, `${pathName}-crt.pem`), certificate) &&
                (!ca || this.writeFile(path_1.default.join(`${storePath}`, `${pathName}-ca.pem`), ca));
        };
        /**
         * In a cluster environment propagate the certificate
         * to the other worker processes via the master process
         */
        this.propagateNewCertificate = (hostname, key, certificate, ca) => {
            if (cluster_1.default.isWorker) {
                this.stats && this.stats.updateCount('CertificatesMessagesSent', 1);
                cluster_1.default.worker.send({
                    messageType: 'certificate',
                    action: 'addCertificate',
                    hostname: hostname,
                    key: key,
                    certificate: certificate,
                    ca: ca
                });
            }
            else {
                delete this.certificates[hostname];
                /**
                 * We only need to load the certificate into the table.
                 * The process that created the certificate should have saved it to a file
                 */
                this.loadCertificate(hostname, key, certificate, ca);
            }
        };
        /**
         * In a cluster environment the master will call this method to propagate the certificate
         */
        this.processMessage = (message) => {
            this.stats && this.stats.updateCount('CertificatesMessagesReceived', 1);
            switch (message.action) {
                case 'addCertificate':
                    delete this.certificates[message.hostname];
                    this.loadCertificate(message.hostname, message.key, message.certificate, message.ca);
                    break;
                default:
                    break;
            }
        };
        /**
         * Read a (set) of PEM file(s) and flatten and unbundle them if they are a bundle
         */
        this.getCertificateData = (pathName, unbundle) => {
            if (pathName) {
                if (Array.isArray(pathName)) {
                    const pathnames = pathName;
                    return pathnames.map((pathname) => {
                        return this.getCertificateData(pathname, unbundle);
                    }).flat();
                }
                else if (this.exists(pathName)) {
                    if (unbundle) {
                        const bundle = this.readFile(pathName, 'utf8');
                        if (bundle) {
                            return this.unbundleCertificate(bundle);
                        }
                        else
                            return null;
                    }
                    else {
                        return this.readFile(pathName, 'utf8');
                    }
                }
            }
        };
        /**
         * Unbundle a PEM string
         */
        this.unbundleCertificate = (bundle) => {
            const lines = bundle ? bundle.trim().split('\n') : [];
            const ca = [];
            let cert = [];
            for (let line of lines) {
                line = line.trim();
                if (!(line.length !== 0)) {
                    continue;
                }
                cert.push(line);
                if (line.match(/-END CERTIFICATE-/)) {
                    ca.push(cert.join('\n'));
                    cert = [];
                }
            }
            return ca;
        };
        /**
         * File system helper functions
         *
         * All operations are sync due to time constraints at startup
         */
        this.exists = (path) => {
            return fs_1.default.existsSync(path);
        };
        this.mkDir = (path, opts = { recursive: true }) => {
            try {
                fs_1.default.mkdirSync(path, opts);
                return true;
            }
            catch (e) {
                return false;
            }
        };
        this.writeFile = (path, data, opts = 'utf8') => {
            try {
                fs_1.default.writeFileSync(path, data, opts);
            }
            catch (e) {
                return false;
            }
            return true;
        };
        this.readFile = (path, encoding = 'utf8') => {
            try {
                return fs_1.default.readFileSync(path, { encoding: encoding, flag: 'r' }).toString();
            }
            catch (e) {
                return null;
            }
        };
        this.certificateStoreRoot = options.certificateStoreRoot;
        this.log = options.log;
        this.stats = options.stats;
    }
}
exports.Certificates = Certificates;
//# sourceMappingURL=certificates.js.map