"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const dns_1 = __importDefault(require("dns"));
/**
 * This is the base class for the DNS update.
 * Specific implementations extend this class to handle
 * the interaction with a particular DNS server
 */
class BaseDNSUpdate {
    constructor(options) {
        /**
         * Add a DNS challenge with verification via DNS server
         */
        this.addDNSChallenge = async (domain, challenge) => {
            /**
             * Remove wildcard characters
             */
            domain = domain.replace(/\*\./g, '');
            this.stats && this.stats.updateCount(`DNSChallengesAddedFor: ${domain}`, 1);
            this.log && this.log.info({ domain: domain }, 'DNS challenge added');
            /**
             * Use the child implementation to write the challenge
             */
            if (await this.addAcmeChallengeToDNS(domain, challenge)) {
                /**
                 * If requested verify the challenge has propagated in the server cluster
                 */
                if (this.dnsNameServer) {
                    /**
                     * Wait for the servers to update
                     */
                    if (await this.waitForNameServerToUpdate(domain, challenge)) {
                        this.stats && this.stats.updateCount('AcmeDNSChallengesWritten', 1);
                        this.log && this.log.info(null, `DNS challenge resolved from ${this.dnsNameServer}`);
                        return true;
                    }
                }
                else {
                    return true;
                }
            }
            else {
                this.stats && this.stats.updateCount(`DnsChallengeWriteFailedFor: ${domain}`, 1);
                this.log && this.log.warn({ domain: domain }, `Write DNS challenge failed`);
                return false;
            }
        };
        /**
         * Remove the DNS challenge via the implementation
         */
        this.removeDNSChallenge = async (domain) => {
            return this.removeAcmeChallengeFromDNS(domain.replace(/\*/g, ''));
        };
        /**
        * Helper method to wait for the DNS servers to update
        */
        this.waitForNameServerToUpdate = async (domain, challenge) => {
            for (let attempts = 0; attempts < this.dnsServerAttempts; ++attempts) {
                try {
                    const resolve = new dns_1.default.promises.Resolver();
                    resolve.setServers([this.dnsNameServer]);
                    const result = await resolve.resolveTxt(`_acme-challenge.${domain}`);
                    const records = [].concat(...result);
                    if (records.indexOf(challenge) >= 0)
                        return true;
                }
                catch (e) {
                    this.stats && this.stats.updateCount(`DNSServerVerificationErrorFor: ${domain}`, 1);
                    this.log && this.log.warn({ domain: domain, error: e }, 'DNS server update verification error');
                }
                finally {
                    await new Promise((resolve) => {
                        setTimeout(resolve, this.dnsServerdelay);
                    });
                }
            }
            return false;
        };
        this.dnsNameServer = options.dnsNameServer;
        this.dnsServerAttempts = options.dnsServerAttempts || 5;
        this.dnsServerdelay = options.dnsServerDelay || 3000;
        this.stats = options.stats;
        this.log = options.log;
    }
}
exports.BaseDNSUpdate = BaseDNSUpdate;
//# sourceMappingURL=dnsUpdate.js.map