"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const axios_1 = __importDefault(require("axios"));
const dnsUpdate_1 = require("./dnsUpdate");
const goDaddyAPIRoot = 'https://api.godaddy.com/v1';
/**
 * This method overrides the empty implmentation in the base class
 */
class GoDaddyDNSUpdate extends dnsUpdate_1.BaseDNSUpdate {
    constructor(options) {
        super(options);
        /**
         * GoDaddy requires us to read all of the TXT records,
         * add our new record, and write back all of the TXT records
         */
        this.addAcmeChallengeToDNS = async (domain, challenge) => {
            this.log && this.log.info({ domain: domain, challenge: challenge }, 'GoDaddy adding ACME challenge');
            const records = await this.getRecordsOfType(domain, 'TXT');
            if (Array.isArray(records)) {
                const txtName = `_acme-challenge`;
                const index = records.findIndex((record) => record.name === txtName);
                if (index >= 0) {
                    records.splice(index, 1);
                }
                records.push({ name: txtName, type: 'TXT', data: challenge, ttl: 600 });
                const result = await this.updateRecordsOfType(domain, 'TXT', records);
                this.log && this.log.debug(result, 'Update results');
                return result;
            }
            return false;
        };
        /**
         * Again we must read all of the TXT records,
         * remove the challenge record, and write the rest back
         */
        this.removeAcmeChallengeFromDNS = async (domain) => {
            this.log && this.log.info({ domain: domain }, 'GoDaddy removing ACME challenge');
            const records = await this.getRecordsOfType(domain, 'TXT');
            if (Array.isArray(records)) {
                const txtName = `_acme-challenge`;
                const index = records.findIndex((record) => record.name === txtName);
                if (index >= 0) {
                    records.splice(index, 1);
                    const result = await this.updateRecordsOfType(domain, 'TXT', records);
                    this.log && this.log.debug(result, 'Update results');
                    return result;
                }
                return true;
            }
            return false;
        };
        /**
         * This is a helper function to read records from the GoDaddy DNS
         */
        this.getRecordsOfType = async (domain, type) => {
            this.log && this.log.info({ domain: domain, type: type }, 'GoDaddy get records of type');
            const url = `${this.buildGoDaddyDomainURL(domain)}/records/${type}`;
            return axios_1.default.get(url, {
                headers: {
                    'Authorization': `sso-key ${this.apiKey}:${this.secret}`,
                }
            })
                .then((response) => {
                if (response.status === 200) {
                    const body = response.data;
                    this.log && this.log.debug({ records: body }, 'GoDaddy result');
                    return body;
                }
                else {
                    this.log && this.log.warn(response.data, `GoDaddy error: ${response.status}`);
                    return [];
                }
            })
                .catch((error) => {
                if (error.response) {
                    this.log && this.log.warn(error.response.data, `GoDaddy read error: ${error.response.status}`);
                    return error;
                }
                else {
                    this.log && this.log.error(error, `GoDaddy read error`);
                    return error;
                }
            });
        };
        this.updateRecordsOfType = async (domain, type, records) => {
            this.log && this.log.debug(records, 'GoDaddy put records of type');
            const url = `${this.buildGoDaddyDomainURL(domain)}/records/${type}`;
            const options = {
                headers: {
                    Authorization: `sso-key ${this.apiKey}:${this.secret}`,
                    'Content-Type': 'application/json',
                }
            };
            return axios_1.default.put(url, records, options)
                .then((response) => {
                if (response.status === 200) {
                    this.log && this.log.info(null, 'GoDaddy update success');
                    return true;
                }
                else {
                    this.log && this.log.warn(response.data, `GoDaddy update error: ${response.status}`);
                    return false;
                }
            })
                .catch((error) => {
                if (error.response) {
                    this.log && this.log.warn(error.response.data, `GoDaddy update error: ${error.response.status}`);
                    return false;
                }
                else {
                    this.log && this.log.error(error, `GoDaddy update error`);
                    return false;
                }
            });
        };
        this.buildGoDaddyDomainURL = (domain) => {
            return `${goDaddyAPIRoot}/domains/${domain}`;
        };
        this.apiKey = options.APIKey;
        this.secret = options.secret;
    }
}
exports.GoDaddyDNSUpdate = GoDaddyDNSUpdate;
//# sourceMappingURL=goDaddyDNSUpdate.js.map