/// <reference types="node" />
import httpProxy from 'http-proxy';
import http from 'http';
import https from 'https';
import net from 'net';
import { Certificates, CertificateOptions } from './certificates';
import { BaseLetsEncryptClient, BaseLetsEncryptOptions } from './letsEncrypt/letsEncrypt';
import { HttpRouter, ExtendedIncomingMessage, RouteRegistrationOptions } from './httpRouter';
import { Logger } from './logger';
import { ProxyUrl } from './util';
import { Statistics } from './statistics';
/**
 * This is the base message type for a clustered environment.
 *
 * Messages from other parts of the system inherit from this
 *
 * Messages are passed both ways from the master to the workers
 */
export interface ClusterMessage {
    messageType: string;
    action: string;
}
/**
 * A minor extension to the http-proxy options to handle NTLM credentials
 */
interface ExtendedProxyOptions extends httpProxy.ServerOptions {
    ntlm?: boolean;
}
/**
 * Startup options for the http server on the front side of the proxy
 */
export interface HttpReverseProxyOptions {
    port?: number;
    host?: string;
    proxyOptions?: ExtendedProxyOptions;
    httpsOptions?: HttpsServerOptions;
    clustered?: boolean | number;
    letsEncryptOptions?: BaseLetsEncryptOptions;
    preferForwardedHost?: boolean;
    log?: Logger | null;
    stats?: Statistics;
}
/**
 * Startup options for the https server on the front side of the reverse proxy
 */
export interface HttpsServerOptions {
    port?: number;
    certificates: Certificates | CertificateOptions;
    host?: string;
    keyFilename?: string;
    certificateFilename?: string;
    caFilename?: string;
    httpsServerOptions?: https.ServerOptions;
}
/**
 * The table of routers. Each router is identified by the inbound host name
 */
declare type HttpRouters = {
    [host: string]: HttpRouter;
};
/**
 * The main Http(s) Reverse Proxy implementation
 */
export declare class HttpReverseProxy {
    protected options: HttpReverseProxyOptions;
    protected isMaster: boolean;
    protected proxy: httpProxy;
    protected server: http.Server;
    protected httpsOptions: HttpsServerOptions;
    protected httpsServer: https.Server;
    protected certificates: Certificates;
    protected letsEncrypt: BaseLetsEncryptClient;
    protected routers: HttpRouters;
    protected log: Logger;
    protected stats: Statistics;
    /**
     * The constructor is passed a set of http options.
     * If no options are passed the default options are used.
     *
     * The second parameter is the type of the Lets Encrypt implementation to use.
     * The default is LetsEncryptUsingAcmeClient
     */
    constructor(options?: HttpReverseProxyOptions, letsEncrypt?: typeof BaseLetsEncryptClient);
    /**
     * Helper method to manage premature worker exit
     */
    private handleWorkerExit;
    /**
     * Helper method to keep track of the workers who are active
     */
    private handleWorkerOnline;
    /**
     * Helper Method to handle messages from the workers
     *
     * Messages are used for logging and the state of the statistics, certificates, and challenges in sync
     */
    private handleWorkerMessage;
    /**
     * The cluster master process. Provides forking, monitoring and messaging for the workers
     */
    private runClusterMaster;
    /**
     * Helper method for the workers to process messages from the master
     */
    private handleServerMessage;
    /**
     *  The primary process of the reverse proxy server
    */
    private runServer;
    /**
     * Add a new route to the system. If this is for a new host create a new HttpRouter.
     *
     * Add the targets to the router
     */
    addRoute: (from: string | Partial<ProxyUrl>, to: string | ProxyUrl | (string | ProxyUrl)[], registrationOptions?: RouteRegistrationOptions) => HttpReverseProxy;
    /**
     * Remove route(s) from an HttpRouter
     */
    removeRoute: (from: string | Partial<URL>, to?: string | ProxyUrl | (string | ProxyUrl)[]) => HttpReverseProxy;
    /**
     * Start the http-proxy service
     */
    protected createProxyServer(proxyOptions: ExtendedProxyOptions): httpProxy;
    /**
     * Start the http server
     */
    protected setupHttpServer: (options: HttpReverseProxyOptions) => http.Server;
    /**
     * Start the https server
     */
    protected setupHttpsServer: (httpsOptions: HttpsServerOptions) => https.Server;
    /**
     * If you are behind another proxy you might want to use the forwarded host name
     */
    protected getInboundHostname: (req: http.IncomingMessage) => string;
    /**
     * Handle the request to upgrade to websockets.
     * Most of the work is handled by http-proxy
     */
    protected websocketsUpgrade: (req: ExtendedIncomingMessage, socket: net.Socket, head: Buffer) => void;
    /**
     * Close down the system gracefully
     *
     * We do not have a handle to the Statistics server here.
     *
     * If you are using a statistics server you must close that in your main code
     */
    close: () => void;
}
export {};
