/// <reference types="node" />
import http from 'http';
import httpProxy from 'http-proxy';
import { Certificates } from './certificates';
import { Logger } from './logger';
import { BaseLetsEncryptClient } from './letsEncrypt/letsEncrypt';
import { ProxyUrl } from './util';
import { Route } from './route';
import { Statistics } from './statistics';
/**
 * LetsEncrypt options when a route is registered. These options take effect
 * on the first route that is registered under a host name.
 *
 * Subsequent calls to addRoute for the same host name with
 * different LetsEncrypt options will be silently ignored.
 */
export interface RegistrationLetsEncryptOptions {
    email: string;
    production?: boolean;
    renewWithin?: number;
    forceRenew?: boolean;
}
/**
 * LetsEncrypt options when a route has an https server. These options take effect
 * on the first route that is registered under a host name.
 *
 * Subsequent calls to addRoute for the same host name with
 * different LetsEncrypt options will be silently ignored.
 */
export interface RegistrationHttpsOptions {
    redirectToHttps: boolean;
    keyFilename?: string;
    certificateFilename?: string;
    caFilename?: string;
    letsEncrypt?: RegistrationLetsEncryptOptions;
}
/**
 * The complete set of registration options for a route/host
 */
export interface RouteRegistrationOptions {
    https?: RegistrationHttpsOptions;
    secureOutbound?: boolean;
    useTargetHostHeader?: boolean;
    stats?: Statistics;
}
/**
 * A minor extension to the http IncommingMessage.
 * Used for a littel internal bookkeepping
 */
export interface ExtendedIncomingMessage extends http.IncomingMessage {
    host: string;
    originalUrl: string;
}
/**
 * The set of options for the HttpRouter.
 * These a normally set by the httpReverseProxy
 */
export interface HttpRouterOptions {
    proxy: httpProxy;
    certificates?: Certificates;
    https?: RegistrationHttpsOptions;
    redirectPort?: number;
    letsEncrypt?: BaseLetsEncryptClient;
    log?: Logger;
    stats?: Statistics;
}
/**
 * The HttpRouter. This handles most of the heavy lifting in the routing process
 */
export declare class HttpRouter {
    protected hostname: string;
    protected proxy: httpProxy;
    protected certificates: Certificates;
    protected https: RegistrationHttpsOptions;
    protected letsEncrypt: BaseLetsEncryptClient;
    protected log: Logger;
    protected stats: Statistics;
    protected redirectPort: number;
    protected routes: Route[];
    private certificateTimer;
    protected certificateFailureCount: number;
    protected letEncryptBusy: boolean;
    /**
     * A new httpRouter requires a hostname and minimal options
     */
    constructor(hostname: string, options: HttpRouterOptions);
    /**
     * Add route consisting of the inbound url and the outbound targets
     */
    addRoute: (from: Partial<URL>, to: string | ProxyUrl | (string | ProxyUrl)[], options?: RouteRegistrationOptions) => this;
    /**
     * Remove a route from the list of targets
     */
    removeRoute: (from: Partial<ProxyUrl>, to?: string | ProxyUrl | (string | ProxyUrl)[]) => this;
    /**
     * Helper method to check we have routes
     */
    noRoutes: () => boolean;
    closeRoute: () => void;
    /**
     * Handle routing of http requests
     */
    routeHttp: (req: ExtendedIncomingMessage, res: http.ServerResponse) => void;
    /**
     * Handle routing https requests
     */
    routeHttps: (req: ExtendedIncomingMessage, res: http.ServerResponse) => void;
    /**
     * Helper method to determine if an http request should be redirected to https
     *
     * If we have https options and they specify http requests should be redirected
     *
     * Then we only need to make sure this is not a request for a LetsEncrypt challenge verification
     */
    private shouldRedirectToHttps;
    /**
     * Helper method to redirect the http request to https
     */
    private redirectToHttps;
    /**
     * Set up the certificateTimer to fire when the Letsencrypt certificate gets too old
     *
     * When the timer fires, a new certificate is requested
     */
    private setCertificateExpiration;
    /**
     * Set up the certificates for this route/hostname
     */
    protected setupCertificates: (options: RegistrationHttpsOptions) => Promise<void>;
    /**
     * Dig through the Routes to find an appropriate Route and target
     */
    getTarget: (req: ExtendedIncomingMessage) => ProxyUrl;
    /**
     * For Mocha testing only
     */
    getTestTarget: (url: string) => ProxyUrl;
    /**
     * Search the Routes for one that matches the inbound request
     */
    protected resolve: (url?: string) => Route;
}
