"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const util_1 = require("./util");
const route_1 = require("./route");
/**
 * Default options for the addRoute with no options
 */
const defaultRegistrationOptions = {
    https: null,
    useTargetHostHeader: false
};
const ONE_DAY = 60 * 60 * 24 * 1000;
const ONE_MONTH = ONE_DAY * 30;
/**
 * The HttpRouter. This handles most of the heavy lifting in the routing process
 */
class HttpRouter {
    /**
     * A new httpRouter requires a hostname and minimal options
     */
    constructor(hostname, options) {
        /**
         * Add route consisting of the inbound url and the outbound targets
         */
        this.addRoute = (from, to, options = defaultRegistrationOptions) => {
            /**
             * There must be a valid source url and target url(s)
             */
            if (!(from = util_1.makeUrl(from)) || !to || (Array.isArray(to) && to.length === 0)) {
                throw Error('Cannot add a new route with invalid "from" or "to"');
            }
            /**
             * Get the source (inbound) path
             */
            const pathname = from.pathname;
            /**
             * See if we already have a route with that path
             */
            let route = this.routes.find((value) => value.path === pathname);
            /**
             * If not, add a new Route
             */
            if (!route) {
                route = new route_1.Route(pathname, this.log, this.stats);
                this.routes.push(route);
                this.stats && this.stats.updateCount(`ActivePathsFor:${this.hostname}`, 1);
                this.stats && this.stats.updateCount(`PathsAddedFor:${this.hostname}`, 1);
            }
            /**
             * Add the targets to the route
             */
            route.addTargets(to, options);
            /**
             * Perform a sanity check on the resulting route
             */
            if (route.noTargets()) {
                this.routes = this.routes.filter((value) => value !== route);
                this.stats && this.stats.updateCount(`ActivePathsFor:${this.hostname}`, -1);
                this.stats && this.stats.updateCount(`PathsRemovedFor:${this.hostname}`, 1);
                throw Error('Cannot add a new route with invalid "from" or "to"');
            }
            /**
             * Sort routes -- longer routes first
             */
            this.routes = this.routes.sort((routeA, routeB) => routeB.path.length - routeA.path.length);
            return this;
        };
        /**
         * Remove a route from the list of targets
         */
        this.removeRoute = (from, to) => {
            if (!(from = util_1.makeUrl(from))) {
                /**
                 * Nothiing to do
                 */
                return this;
            }
            /**
             * Get the source (inbound) path
             */
            const pathname = from.pathname;
            /**
             * See if we have a Route with that path
             */
            const route = this.routes.find((route) => route.path === pathname);
            if (route) {
                /**
                 * Remove the matching targets
                 */
                route.removeTargets(to);
                /**
                 * If the Route has no targets, remove the Route
                 */
                if (route.noTargets()) {
                    this.routes = this.routes.filter((value) => value !== route);
                    this.stats && this.stats.updateCount(`ActivePathsFor:${this.hostname}`, -1);
                    this.stats && this.stats.updateCount(`PathsRemovedFor:${this.hostname}`, 1);
                }
            }
            this.log && this.log.info({ from: from, to: to }, 'Unregistered a route');
            return this;
        };
        /**
         * Helper method to check we have routes
         */
        this.noRoutes = () => {
            return this.routes.length === 0;
        };
        this.closeRoute = () => {
            if (this.certificateTimer) {
                this.certificateTimer.clearTimer();
            }
        };
        /**
         * Handle routing of http requests
         */
        this.routeHttp = (req, res) => {
            /**
             * Get a target matching the route
             */
            const target = this.getTarget(req);
            this.stats && this.stats.updateCount(`HttpRouteRequestsFor:${this.hostname}`, 1);
            /**
             * If we have a matching target we can pass the request on to http-proxy
             */
            if (target) {
                /**
                 * Check if this http request should be redirected to https
                 */
                if (this.shouldRedirectToHttps(target)) {
                    /**
                     * If so, inform the browser
                     */
                    this.redirectToHttps(req, res);
                }
                else {
                    try {
                        /**
                         * Pass the request on to http-proxy. Target.secure is set by RouteRegistrationOptions.secureOutbound
                         */
                        this.proxy.web(req, res, { target: target, secure: target.secure }, 
                        /**
                         * If it fails, let the client know we are lost
                         */
                        (error, req, res) => {
                            util_1.respondNotFound(req, res);
                            this.stats && this.stats.updateCount(`HttpProxyFailuresFor:${this.hostname}`, 1);
                        });
                    }
                    catch (err) {
                        util_1.respondNotFound(req, res);
                        this.stats && this.stats.updateCount(`HttpProxyFailuresFor:${this.hostname}`, 1);
                    }
                }
            }
            else {
                util_1.respondNotFound(req, res);
                this.stats && this.stats.updateCount(`HttpRouteFailuresFor:${this.hostname}`, 1);
            }
        };
        /**
         * Handle routing https requests
         */
        this.routeHttps = (req, res) => {
            /**
             * Get the target matching the route
             */
            const target = this.getTarget(req);
            this.stats && this.stats.updateCount(`HttpsRouteRequestsFor:${this.hostname}`, 1);
            /**
             * If we have a matching target we can pass the request on to http-proxy
             */
            if (target) {
                try {
                    // req.headers.host = target.host
                    this.proxy.web(req, res, { target: target, secure: target.secure }, (error, req, res) => {
                        util_1.respondNotFound(req, res);
                        this.stats && this.stats.updateCount(`HttpsProxyFailuresFor:${this.hostname}`, 1);
                    });
                }
                catch (err) {
                    util_1.respondNotFound(req, res);
                    this.stats && this.stats.updateCount(`HttpsProxyFailuresFor:${this.hostname}`, 1);
                }
            }
            else {
                util_1.respondNotFound(req, res);
                this.stats && this.stats.updateCount(`HttpsRouteFailuresFor:${this.hostname}`, 1);
            }
        };
        /**
         * Helper method to determine if an http request should be redirected to https
         *
         * If we have https options and they specify http requests should be redirected
         *
         * Then we only need to make sure this is not a request for a LetsEncrypt challenge verification
         */
        this.shouldRedirectToHttps = (target) => {
            return this.https
                && this.https.redirectToHttps
                && (!this.letsEncrypt || target.href != this.letsEncrypt.href);
        };
        /**
         * Helper method to redirect the http request to https
         */
        this.redirectToHttps = (req, res) => {
            req.url = req.originalUrl || req.url; // Get the original url since we are going to redirect.
            const targetPort = this.redirectPort;
            /**
             * Set up the redirect to use the same host and url
             */
            const hostname = this.hostname + (targetPort ? ':' + targetPort : '');
            const url = 'https://' + path_1.default.join(hostname, req.url).replace(/\\/g, '/');
            this.log && this.log.info(null, `Redirecting ${path_1.default.join(req.headers.host, req.url)} to ${url}`);
            //
            // We can use 301 for permanent redirect, but its bad for debugging, we may have it as
            // a configurable option.
            //
            res.writeHead(302, { Location: url });
            res.end();
            this.stats && this.stats.updateCount(`HttpRedirectsFor:${this.hostname}`, 1);
        };
        /**
         * Set up the certificateTimer to fire when the Letsencrypt certificate gets too old
         *
         * When the timer fires, a new certificate is requested
         */
        this.setCertificateExpiration = (expiresOn, renewWithin = ONE_MONTH) => {
            /**
             * If the timer is already set, ignore this request
             */
            if (this.certificateTimer) {
                return;
            }
            let expiration = expiresOn ? expiresOn.valueOf() - Date.now() - renewWithin : 0;
            /**
             * If we are already overdue, wait at least a minute
             */
            if (expiration <= 0) {
                expiration = 60000;
            }
            this.certificateTimer = util_1.setLongTimeout(async () => {
                /**
                 * Clear the certificate timer. It will be set by the request for a certificate
                 */
                this.certificateTimer = null;
                /**
                 * If we get the certificate, we are done
                 */
                if (await this.letsEncrypt.getLetsEncryptCertificate(this.hostname, this.https.letsEncrypt.production, this.https.letsEncrypt.email, true, this.setCertificateExpiration)) {
                    this.certificateFailureCount = 0;
                }
                else {
                    /**
                     * Do not allow unlimited failures
                     */
                    if (++this.certificateFailureCount > 5) {
                        this.log && this.log.error({ hostname: this.hostname }, 'LetsEncrypt get certificate aborted');
                        this.stats && this.stats.updateCount(`LetsEncryptCertificateAborted: ${this.hostname}`, 1);
                    }
                    else {
                        this.log && this.log.error({ hostname: this.hostname }, 'LetsEncrypt get certificate failed');
                        this.stats && this.stats.updateCount(`LetsEncryptCertificateFailed: ${this.hostname}`, 1);
                        /**
                         * If we have a certificate (expiresOn not null), try again tomorrow
                         * Otherwise we will try again in 10 seconds.
                         */
                        this.setCertificateExpiration(expiresOn, ONE_MONTH - this.certificateFailureCount * ONE_DAY);
                    }
                }
            }, expiration);
        };
        /**
         * Set up the certificates for this route/hostname
         */
        this.setupCertificates = async (options) => {
            if (!this.certificates) {
                throw Error('Cannot register https routes without certificate option');
            }
            /**
             * If we do not have a certificate see if we can request one
             */
            if (!this.certificates.getCertificate(this.hostname)) {
                if ('object' === typeof options) {
                    if (options.keyFilename || options.certificateFilename || options.caFilename) {
                        /**
                         * We are using permanent certificates for this route/host
                         */
                        if (this.certificates.loadCertificateFromFiles(this.hostname, options.keyFilename, options.certificateFilename, options.caFilename)) {
                            this.stats && this.stats.updateCount(`CertificatesLoadedFor:${this.hostname}`, 1);
                        }
                        else {
                            this.stats && this.stats.updateCount(`CertificatesMissing:${this.hostname}`, 1);
                            this.log && this.log.error({ hostname: this.hostname }, 'Missing cerrtificate files');
                        }
                    }
                    else if (options.letsEncrypt) {
                        /**
                         * Set up for a temporary certificate
                         */
                        if (!this.letsEncrypt) {
                            /**
                             * We have a configuration issue
                             */
                            this.log && this.log.error({ hostname: this.hostname }, 'Missing LetsEncrypt in router configuration');
                            return;
                        }
                        /**
                         * Test to see if another route is already working on it.
                         * Requesting a certificate is an asynchronous process.
                         */
                        if (this.letEncryptBusy) {
                            return;
                        }
                        this.letEncryptBusy = true;
                        this.log && this.log.info(null, `Getting Let's Encrypt certificates for ${this.hostname}`);
                        this.stats && this.stats.updateCount(`CertificateRequestsFor:${this.hostname}`, 1);
                        /**
                         * See if LetsEncrypt can get us a certificate
                         */
                        if (!await this.letsEncrypt.getLetsEncryptCertificate(this.hostname, options.letsEncrypt.production, options.letsEncrypt.email, options.letsEncrypt.forceRenew, this.setCertificateExpiration)) {
                            /**
                             * If we fail schedule a retry in 60 seconds
                             */
                            this.log && this.log.warn({ hostname: this.hostname }, 'Initial certificate request failed');
                            this.stats && this.stats.updateCount(`FailedCertificateRequestsFor: ${this.hostname}`, 1);
                            this.setCertificateExpiration(new Date(Date.now() + 120000), 60000);
                        }
                        this.letEncryptBusy = false;
                    }
                }
                else {
                    // Trigger the use of the default certificates.
                    this.certificates.addCertificate(this.hostname, null);
                }
            }
        };
        /**
         * Dig through the Routes to find an appropriate Route and target
         */
        this.getTarget = (req) => {
            const url = req.url;
            /**
             * Try to resolve the route
             */
            const route = this.resolve(url);
            if (!route) {
                this.stats && this.stats.updateCount(`TargetFailuresFor:${this.hostname}`, 1);
                this.log && this.log.warn({ host: req.headers.host, url: url }, 'no valid route found for given source');
                return null;
            }
            /**
             * Fix up the source path
             */
            const pathname = route.path;
            if (pathname.length > 1) {
                /**
                 * Save the original url for later.
                 * Remove the prefix specified in Route from the request Url
                 * Leaving just the remaining part
                  * Route.url === '/abc'
                  * req.url === '/abc/def'
                  * req.url = '/def'
                */
                req.originalUrl = url; // save original url
                req.url = url.substr(pathname.length) || '/';
            }
            /**
             * Get the next target in the route
             */
            const target = route.nextTarget();
            /**
             * Fix request url if target pathname is specified.
             * Append the remainder of the req.url to the target path
               * Route.url === '/abc'
               * req.url === '/abc/def'
               * target.url === '/123'
               * req.url = '/123/def'
             */
            if (target.pathname) {
                req.url = path_1.default.join(target.pathname, req.url).replace(/\\/g, '/');
            }
            /**
             * Host headers are passed through from the source by default
             * We may want to use the host header of the target instead
             * specifically if we have proxies behind us
             * or servers that check the host name matches their own
             */
            if (target.useTargetHostHeader === true) {
                req.headers.host = target.host;
            }
            this.stats && this.stats.updateCount(`TargetRequestsFor:${this.hostname}${pathname} to ${target.host}${target.pathname}`, 1);
            this.log && this.log.info(null, `Proxying ${req.headers.host}${pathname} to ${target.href}${target.pathname}`);
            return target;
        };
        /**
         * For Mocha testing only
         */
        this.getTestTarget = (url) => {
            const route = this.resolve(url);
            const Url = util_1.makeUrl(url);
            if (!route) {
                this.log && this.log.warn({ host: Url.host, url: url }, 'no valid route found for given source');
                return null;
            }
            const pathname = route.path;
            const target = route.nextTarget();
            //
            // Fix request url if targetname specified.
            //
            if (target.pathname) {
                url = path_1.default.join(target.pathname, url).replace(/\\/g, '/');
            }
            this.log && this.log.info(null, `Testing ${Url.host}/${url} to ${target.href}/${url}`);
            return target;
        };
        /**
         * Search the Routes for one that matches the inbound request
         */
        this.resolve = (url) => {
            this.stats && this.stats.updateCount(`RouteResolutionsFor:${this.hostname}`, 1);
            url = url || '/';
            /**
             * This is a hack.
             *
             * If we are looking for the LetsEncrypt server, pass back an uregistered Route
             */
            if (this.letsEncrypt && /^\/.well-known\/acme-challenge\//.test(url)) {
                this.log && this.log.info(null, `Adding LetsEncrypt Route ${url}`);
                this.stats && this.stats.updateCount(`AcmeRequestsFor:${this.hostname}`, 1);
                return new route_1.Route('/').addTargets(this.letsEncrypt.href, {});
            }
            /**
             * Look for the first Route that matches the inbound URL
             *
             * Since the Routes are sorted with the longest first, we will find the most accurate match
             */
            return this.routes.find((route) => route.path === '/' || util_1.startsWith(url, route.path)) || null;
        };
        this.hostname = hostname;
        this.certificates = options.certificates;
        this.proxy = options.proxy;
        this.https = options.https;
        this.letsEncrypt = options.letsEncrypt;
        this.redirectPort = options.redirectPort;
        /**
         * Routes are stored iin a sorted array with the longest inbound url first.
         * This makes sure a request for server1.test.com/api comes before server1.test.api/
         */
        this.routes = [];
        this.log = options.log;
        this.stats = options.stats;
        /**
         * The certificate timer is used to request a new certificate before the current one expires
         */
        this.certificateTimer = null;
        /**
         * Too many errors requesting a new certificate will abort the process
         */
        this.certificateFailureCount = 0;
        /**
         * If https is specified in the options, set up the certificates
         */
        if (options.https) {
            this.setupCertificates(options.https);
        }
    }
}
exports.HttpRouter = HttpRouter;
//# sourceMappingURL=httpRouter.js.map