/// <reference types="node" />
import http from 'http';
import { Certificates } from "../certificates";
import { BaseDNSUpdate } from "../dns/dnsUpdate";
import { Logger } from "../logger";
import { ClusterMessage } from '../httpReverseProxy';
import { Statistics } from '../statistics';
/**
 * The base Let's Encrypt options
 *
 * The base class instantiates an http server to respond to the Http-01 challenges
 * The port and host are used by the server
 *
 * The dnsChallenge is the interface to a particular DNS server
 * when you need to support a dns-01 challenge.
 *
 * The dnsNameServer is optional. If specified it is the ip address of the DNS nameserver.
 * When specified the DNS challenge will query the nameserver to make sure the
 * update to the TXT record has propagated to the server specified
 */
export interface BaseLetsEncryptOptions {
    port?: number;
    host?: string;
    certificates?: Certificates;
    dnsChallenge?: BaseDNSUpdate;
    log?: Logger;
    stats?: Statistics;
}
/**
 * The http-01 challenge table is a simple object with host name and modified tokens as keys
 * and the challenge response as the value
 */
interface HttpChallengeTable {
    [hostAndToken: string]: string;
}
/**
 * For clustered environment the challenges are distributed to all of the
 * cluster workers through messages handled by the cluster master
 */
export interface LetsEncryptMessage extends ClusterMessage {
    host: string;
    token: string;
    keyAuthorization?: string;
}
/**
 * THis is the base class for the specific implentation of
 * the Let's Encrypt Acme challenge handlers.
 *
 * It instantiates the Http server to handle the http01 challenges
 */
export declare class BaseLetsEncryptClient {
    protected certificates: Certificates;
    protected networkInterface: string;
    protected port: number;
    protected httpServer: http.Server;
    protected dnsChallenge: BaseDNSUpdate;
    protected log: Logger;
    protected stats: Statistics;
    protected outstandingChallenges: HttpChallengeTable;
    constructor(options: BaseLetsEncryptOptions);
    /**
     * This method should be overwritten by new implementations.
     * See the letsEncryptAcmeClient for details
     */
    protected getNewCertificate: (host: string, production: boolean, email: string) => Promise<boolean>;
    get href(): string;
    get serverInterface(): string;
    get serverPort(): number;
    /**
     * Start an Http server to process the challenges from Let's Encrypt
     */
    private setupLetsEncryptServer;
    /**
     * Shut down the server to allow for a clean shutdown of the proxy
     */
    close: () => void;
    /**
     * This is the entry point for requesting a new certificate.
     *
     * The production parameter should be false during early testing. Too many requests to the
     * production servers will disable the account for a period of time.
     *
     * ForceRenew will ignore the expiration of an existing certificate and request a new certificate.
     *
     * setExpiration is a callback to allow the calling code to monitor the expiration date of the
     * certificate and request a new one before it expires
     */
    getLetsEncryptCertificate: (hostname: string, production: boolean, email: string, forceRenew?: boolean, setExpiration?: (expiresOn: Date) => void) => Promise<boolean>;
    /**
     * Helper function to turn the hostname and token into a key for the certiciate table
     */
    private makeKey;
    /**
     * The getNewCertificate method will use this method to add a challenge to the table.
     *
     * If we are working in a cluster the update is sent to the master process to be routed to
     * all of the worker processes
     */
    protected addChallenge: (host: string, token: string, keyAuthorization: string) => void;
    /**
     * Once a challenge is complete, remove it.
     *
     * If we are working in a cluster the update is sent to the master process to be routed to
     * all of the worker processes
     */
    protected removeChallenge: (host: string, token: string) => void;
    /**
     * If we are working in a cluster messages will arrive from the server to update
     * the table. Process the Add and Remove messages. This method is called by the master
     * process with any challenge messages it receives
     */
    processMessage: (message: LetsEncryptMessage) => void;
}
export {};
