"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const http_1 = __importDefault(require("http"));
const cluster_1 = __importDefault(require("cluster"));
/**
 * THis is the base class for the specific implentation of
 * the Let's Encrypt Acme challenge handlers.
 *
 * It instantiates the Http server to handle the http01 challenges
 */
class BaseLetsEncryptClient {
    constructor(options) {
        /**
         * This method should be overwritten by new implementations.
         * See the letsEncryptAcmeClient for details
         */
        this.getNewCertificate = async (host, production, email) => {
            return false;
        };
        /**
         * Start an Http server to process the challenges from Let's Encrypt
         */
        this.setupLetsEncryptServer = () => {
            const server = http_1.default.createServer();
            server.on('request', (req, res) => {
                this.log && this.log.info(null, `LetsEncypt validate url: ${req.url}`);
                /**
                 * The server will be accessable from the outside
                 * so care must be taken to only respond to valid Urls and tokens
                 */
                const validPath = /^\/.well-known\/acme-challenge\//.test(req.url);
                if (validPath) {
                    /**
                     * trim white space,
                     * strip trailing /,
                     * split into array on /,
                     * take last item,
                     * remove invalid identifier characters
                     */
                    const token = req.url.trim().replace(/$\//, '').split('/').pop().replace(/\W-/g, '');
                    this.log && this.log.info({ token: token }, `LetsEncrypt validating challenge`);
                    if (token && this.outstandingChallenges[`${req.headers.host}_${token}`]) {
                        // send the key corresponding to the token from the outstanding challenges
                        res.writeHead(200);
                        res.end(this.outstandingChallenges[`${req.headers.host}_${token}`]);
                        return;
                    }
                }
                // respond with error if missing challenge path, token, or token is not in in outstanding challenges
                res.writeHead(404, { "Content-Type": "text/plain" });
                res.end("404 Not Found\n");
            });
            server.on('listening', () => {
                const serverAddress = server.address();
                this.stats && this.stats.updateCount('LetsEncryptServersStarted', 1);
                this.stats && this.stats.updateCount('LetsEncryptServersRunning', 1);
                this.log && this.log.info(serverAddress, `LetsEncrypt server listening to Http requests`);
            });
            server.listen(this.port, this.networkInterface);
            return server;
        };
        /**
         * Shut down the server to allow for a clean shutdown of the proxy
         */
        this.close = () => {
            this.stats && this.stats.updateCount('LetsEncryptServersRunning', -1);
            this.httpServer && this.httpServer.close();
        };
        /**
         * This is the entry point for requesting a new certificate.
         *
         * The production parameter should be false during early testing. Too many requests to the
         * production servers will disable the account for a period of time.
         *
         * ForceRenew will ignore the expiration of an existing certificate and request a new certificate.
         *
         * setExpiration is a callback to allow the calling code to monitor the expiration date of the
         * certificate and request a new one before it expires
         */
        this.getLetsEncryptCertificate = async (hostname, production, email, forceRenew, setExpiration) => {
            /**
             * Initiate the callback if required
             */
            const updateExpiration = () => {
                if ('function' === typeof setExpiration) {
                    const certificateData = this.certificates.getCertificateInformation(hostname);
                    setExpiration(certificateData.expiresOn);
                }
            };
            this.stats && this.stats.updateCount('LetsEncryptCertificateRequests', 1);
            /**
             * First try to load the certificate from the store
             */
            if (this.certificates.loadCertificateFromStore(hostname) && !forceRenew) {
                this.stats && this.stats.updateCount('LetsEncryptCertificatesLoaded', 1);
                updateExpiration();
                return true;
            }
            /**
             * If we are operating in a cluster then there are many of these processes running.
             * Care must be taken to limit the number of requests for a certificate from multiple
             * processes.
             */
            if (cluster_1.default.isWorker) {
                /**
                 * If we have a certificate and we are forcing a new one, then do a little bookkeeping to
                 * determine if another process beat us to it.
                 */
                let certificateData = this.certificates.getCertificateInformation(hostname);
                const expiration = certificateData && certificateData.expiresOn;
                /**
                 * set a random delay to avoid startup race condition
                 * with all of the processes trying to aquire expired or missing
                 * certificates
                 */
                await new Promise((resolve) => {
                    setTimeout(resolve, Math.random() * 60000);
                });
                /**
                 * If the certificate expiration has changed, another process has already updated it.
                 */
                if (expiration &&
                    this.certificates.getCertificateInformation(hostname) &&
                    this.certificates.getCertificateInformation(hostname).expiresOn > expiration) {
                    this.stats && this.stats.updateCount('LetsEncryptCertificatesRemotelyResolved', 1);
                    updateExpiration();
                    return true;
                }
            }
            /**
             * Request a new certificate from the implementation of this instance
             */
            this.stats && this.stats.updateCount('LetsEncryptNewCertificateRequests', 1);
            this.log && this.log.warn({ hostname: hostname }, 'Getting new LetsEncrypt certificate');
            if (await this.getNewCertificate(hostname, production, email)) {
                this.log && this.log.warn({ hostname: hostname }, 'New LetsEncrypt certificate received');
                updateExpiration();
                return true;
            }
            return false;
        };
        /**
         * Helper function to turn the hostname and token into a key for the certiciate table
         */
        this.makeKey = (hostname, token) => {
            return `${hostname}_${token.replace(/\W-/g, '')}`;
        };
        /**
         * The getNewCertificate method will use this method to add a challenge to the table.
         *
         * If we are working in a cluster the update is sent to the master process to be routed to
         * all of the worker processes
         */
        this.addChallenge = (host, token, keyAuthorization) => {
            if (cluster_1.default.isWorker) {
                cluster_1.default.worker.send({
                    messageType: 'letEncrypt',
                    action: 'addChallenge',
                    host: host,
                    token: token,
                    keyAuthorization: keyAuthorization
                });
            }
            else {
                this.stats && this.stats.updateCount('LetsEncryptHttpChallengesAdded', 1);
                this.outstandingChallenges[this.makeKey(host, token)] = keyAuthorization;
            }
        };
        /**
         * Once a challenge is complete, remove it.
         *
         * If we are working in a cluster the update is sent to the master process to be routed to
         * all of the worker processes
         */
        this.removeChallenge = (host, token) => {
            if (cluster_1.default.isWorker) {
                this.stats && this.stats.updateCount('LetsEncryptMessagesSent', 1);
                cluster_1.default.worker.send({
                    messageType: 'letEncrypt',
                    action: 'removeChallenge',
                    host: host,
                    token: token
                });
            }
            else {
                this.stats && this.stats.updateCount('LetsEncryptHttpChallengesRemoved', 1);
                delete this.outstandingChallenges[this.makeKey(host, token)];
            }
        };
        /**
         * If we are working in a cluster messages will arrive from the server to update
         * the table. Process the Add and Remove messages. This method is called by the master
         * process with any challenge messages it receives
         */
        this.processMessage = (message) => {
            if (message.messageType === 'letEncrypt') {
                this.stats && this.stats.updateCount('LetsEncryptMessagesReceived', 1);
                switch (message.action) {
                    case 'addChallenge':
                        this.stats && this.stats.updateCount('LetsEncryptHttpChallengesAdded', 1);
                        this.outstandingChallenges[this.makeKey(message.host, message.token)] = message.keyAuthorization;
                        break;
                    case 'removeChallenge':
                        this.stats && this.stats.updateCount('LetsEncryptHttpChallengesRemoved', 1);
                        delete this.outstandingChallenges[this.makeKey(message.host, message.token)];
                        break;
                    default:
                        break;
                }
            }
        };
        this.certificates = options.certificates;
        this.networkInterface = options.host;
        this.port = options.port || 3000;
        this.httpServer = this.setupLetsEncryptServer();
        this.dnsChallenge = options.dnsChallenge;
        // this.dnsNameServer = options.dnsNameServer
        this.log = options.log;
        this.stats = options.stats;
        this.outstandingChallenges = {};
    }
    get href() {
        return `http://${this.networkInterface || 'localhost'}:${this.port}/.well-known/acme-challenge`;
    }
    get serverInterface() {
        return this.networkInterface;
    }
    get serverPort() {
        return this.port;
    }
}
exports.BaseLetsEncryptClient = BaseLetsEncryptClient;
//# sourceMappingURL=letsEncrypt.js.map