"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const node_forge_1 = __importDefault(require("node-forge"));
const letsEncrypt_1 = require("./letsEncrypt");
/**
 * We set the self signed certificate to expire after three months
 */
const THREE_MONTHS = 90 * 24 * 60 * 60 * 1000;
class LetsEncryptUsingSelfSigned extends letsEncrypt_1.BaseLetsEncryptClient {
    constructor(options) {
        super(options);
        /**
         * This is the method that overrides the base class method
         */
        this.getNewCertificate = async (host, production, email) => {
            /**
             * this was taken straight from the forge example
             */
            const keys = node_forge_1.default.pki.rsa.generateKeyPair(1024);
            const certificate = node_forge_1.default.pki.createCertificate();
            certificate.publicKey = keys.publicKey;
            certificate.serialNumber = '01';
            certificate.validity.notBefore = new Date();
            certificate.validity.notAfter = new Date(new Date().valueOf() + THREE_MONTHS);
            const attrs = [{
                    name: 'commonName',
                    value: host
                }, {
                    name: 'countryName',
                    value: this.country
                }, {
                    shortName: 'ST',
                    value: this.state
                }, {
                    name: 'localityName',
                    value: this.locality
                }, {
                    name: 'organizationName',
                    value: this.organizationName
                }, {
                    shortName: 'OU',
                    value: this.organizationName
                }];
            certificate.setSubject(attrs);
            certificate.setIssuer(attrs);
            certificate.setExtensions([{
                    name: 'basicConstraints',
                    cA: true
                }, {
                    name: 'keyUsage',
                    keyCertSign: true,
                    digitalSignature: true,
                    nonRepudiation: true,
                    keyEncipherment: true,
                    dataEncipherment: true
                }, {
                    name: 'extKeyUsage',
                    serverAuth: true,
                    clientAuth: true,
                    codeSigning: true,
                    emailProtection: true,
                    timeStamping: true
                }, {
                    name: 'nsCertType',
                    client: true,
                    server: true,
                    email: true,
                    objsign: true,
                    sslCA: true,
                    emailCA: true,
                    objCA: true
                }, {
                    name: 'subjectAltName',
                    altNames: [{
                            type: 6,
                            value: `http://${host}`
                        }, {
                            type: 7,
                            ip: '127.0.0.1'
                        }]
                }, {
                    name: 'subjectKeyIdentifier'
                }]);
            // self-sign certificate
            certificate.sign(keys.privateKey);
            // PEM-format keys and cert
            const pem = {
                privateKey: node_forge_1.default.pki.privateKeyToPem(keys.privateKey),
                publicKey: node_forge_1.default.pki.publicKeyToPem(keys.publicKey),
                certificate: node_forge_1.default.pki.certificateToPem(certificate)
            };
            this.stats && this.stats.updateCount('SelfSignedCertificatesRequested', 1);
            this.log && this.log.info(null, 'Certificate created.');
            this.certificates.saveCertificateToStore(host, pem.privateKey, pem.certificate);
            this.certificates.propagateNewCertificate(host, pem.privateKey, pem.certificate);
            return true;
        };
        this.country = options.country;
        this.state = options.state;
        this.locality = options.locality;
        this.organizationName = options.organizationName;
        this.callCount = 0;
    }
}
exports.LetsEncryptUsingSelfSigned = LetsEncryptUsingSelfSigned;
//# sourceMappingURL=letsEncryptUsingSelfSigned.js.map