"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const http_1 = __importDefault(require("http"));
const ws_1 = __importDefault(require("ws"));
const fs_1 = __importDefault(require("fs"));
const cluster_1 = __importDefault(require("cluster"));
const safe_1 = __importDefault(require("colors/safe"));
const util_1 = require("./util");
exports.logLevels = [
    { level: 10, type: 'debug' },
    { level: 20, type: 'trace' },
    { level: 30, type: 'info' },
    { level: 40, type: 'warn' },
    { level: 50, type: 'error' },
    { level: 60, type: 'fatal' },
];
/**
 * Class to handle the individual WebSocket connections
 */
class WebSocketClient {
    constructor(logLevel, ws) {
        this.open = () => {
        };
        this.close = () => {
            this.ws && this.ws.close();
        };
        this.logLevel = logLevel;
        this.ws = ws;
        ws.on('open', this.open);
        /**
         * Handle messages from the websocket client.
         * Client can start or stop the sending of data,
         * change the update interval or set the filters
         */
        ws.on('message', (message) => {
            const data = JSON.parse(message);
            if (data && data.command) {
                switch (data.command) {
                    case 'start':
                        this.open();
                        break;
                    case 'stop':
                        this.close();
                        break;
                    case 'setLogLevel':
                        this.logLevel = Number(data.logLevel);
                        break;
                    default:
                        break;
                }
            }
        });
        ws.on('close', this.close);
    }
    sendLogMessage(level, data, message) {
        if (level >= this.logLevel) {
            const type = exports.logLevels.find((testLevel) => testLevel.level >= level);
            this.ws.send(JSON.stringify({ level: type ? type.type : level, data: data, message: message }));
        }
    }
}
class Logger {
    constructor(options) {
        this.logLevel = 10;
        this.clients = [];
        /**
         * Start the Logging http server
         */
        this.createHttpServer = () => {
            this.httpServer = http_1.default.createServer();
            this.httpServer.on('request', (req, res) => {
                if (this.htmlFilename) {
                    fs_1.default.readFile(this.htmlFilename, (error, data) => {
                        if (!error) {
                            res.write(data.toString('utf8'));
                            res.end();
                        }
                        else {
                            util_1.respondNotFound(req, res);
                        }
                    });
                }
                else {
                    util_1.respondNotFound(req, res);
                }
            });
            this.httpServer.on('upgrade', (req, socket, head) => {
                this.websocketServer.handleUpgrade(req, socket, head, (ws) => {
                    this.websocketServer.emit('connection', ws, req);
                });
            });
            this.httpServer.on('listening', () => {
                console.log(`info: ${JSON.stringify(this.httpServer.address())} Logging server started`);
            });
            this.httpServer.listen(this.port, this.host);
        };
        /**
         * Create the websocket server
         */
        this.createWebsocketServer = () => {
            this.websocketServer = new ws_1.default.Server({ noServer: true });
            /**
             * Handle connections to the websocket server
             */
            this.websocketServer.on('connection', (ws) => {
                this.clients.push(new WebSocketClient(this.logLevel, ws));
                ws.on('close', () => {
                    this.clients = this.clients.filter((client) => client.ws !== ws);
                });
            });
        };
        this.log = (type, color, data, message) => {
            const logLevel = exports.logLevels.find((level) => level.type === type);
            if (!logLevel || logLevel.level >= this.logLevel) {
                if (cluster_1.default.isWorker) {
                    process.send({
                        messageType: 'logging',
                        action: 'log',
                        workerId: this.workerId,
                        type: type,
                        data: data,
                        message: message
                    });
                }
                else {
                    if (data) {
                        console.log(color(`${type}: ${JSON.stringify({ time: new Date().toLocaleString(), ...data })}: ${message}`));
                    }
                    else {
                        console.log(color(`${type}: ${JSON.stringify({ time: new Date().toLocaleString() })} ${message}`));
                    }
                    this.clients.forEach((client) => {
                        client.sendLogMessage(logLevel.level, data, message);
                    });
                }
            }
        };
        this.processMessage = (message) => {
            switch (message.action) {
                case 'log':
                    if (cluster_1.default.isMaster) {
                        try {
                            this[message.type](message.data, message.message);
                        }
                        catch (e) {
                            this.fatal(message, 'Unknown remote message type');
                        }
                    }
                    break;
                default:
                    break;
            }
        };
        this.close = () => {
            this.httpServer && this.httpServer.close();
            this.clients.forEach((client) => {
                client.close();
            });
            this.websocketServer && this.websocketServer.close();
        };
        if (cluster_1.default.isWorker) {
            this.workerId = cluster_1.default.worker.id;
        }
        else {
            this.workerId = 0;
        }
        if (options) {
            this.host = options.host;
            this.port = options.port || 3002;
            this.htmlFilename = options.htmlFilename;
            this.logLevel = options.logLevel || 30;
            if (this.workerId === 0 && (options.host || options.port)) {
                this.createHttpServer();
                this.createWebsocketServer();
            }
        }
    }
    debug(data, message) {
        this.log('debug', safe_1.default.blue, data, message);
    }
    trace(data, message) {
        this.log('trace', safe_1.default.green, data, message);
    }
    info(data, message) {
        this.log('info', safe_1.default.white, data, message);
    }
    warn(data, message) {
        this.log('warn', safe_1.default.yellow, data, message);
    }
    error(data, message) {
        this.log('error', safe_1.default.red, data, message);
    }
    fatal(data, message) {
        this.log('fatal', safe_1.default.bgRed, data, message);
    }
}
exports.Logger = Logger;
//# sourceMappingURL=logger.js.map