"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const http_1 = __importDefault(require("http"));
const fs_1 = __importDefault(require("fs"));
const ws_1 = __importDefault(require("ws"));
const util_1 = require("./util");
/**
 * The statistics server is started with an instance of the Statistics class.
 * This is most often a singleton class created in the root of the proxy startup
 * code. Statistics entries are updated in the Statistics object.
 * The Statistics server sends to current state of the Statistics instance to
 * out to the clients via a websocket interface
 */
class StatisticsServer {
    constructor(options) {
        /**
         * Extract the option information from the statistics options
         */
        this.setOptions = (options) => {
            if (options) {
                this.stats = options.stats;
                this.port = 3001;
                this.htmlFilename = './public/statisticsAndLoggingPage.html';
                this.updateInterval = 5000;
                this.htmlFilename = options.htmlFilename || this.htmlFilename;
                this.log = options.log;
                if (options.http) {
                    this.host = options.http.host;
                    this.port = options.http.port || this.port;
                }
                if (options.websocket) {
                    this.updateInterval = options.websocket.updateInterval || this.updateInterval;
                    this.filter = options.websocket.filter;
                }
            }
        };
        /**
         * The statistics server can be created and started in a single step or the
         * server can be started later in the overall startup process with additional/
         * changed options
         */
        this.start = (options) => {
            this.stop();
            this.setOptions(options);
            this.createHttpServer();
            this.createWebsocketServer();
        };
        /**
         * Stop the statistics server. It can be restarted again.
         */
        this.stop = () => {
            this.httpServer && this.httpServer.close();
            if (this.websocketServer) {
                this.websocketServer.clients.forEach((client) => {
                    client.close();
                });
                this.websocketServer.close();
            }
        };
        /**
         * Start the Statistics server http server
         */
        this.createHttpServer = () => {
            this.httpServer = http_1.default.createServer();
            this.httpServer.on('request', (req, res) => {
                fs_1.default.readFile(this.htmlFilename, (error, data) => {
                    if (!error) {
                        res.write(data.toString('utf8'));
                        res.end();
                    }
                    else {
                        util_1.respondNotFound(req, res);
                    }
                });
            });
            this.httpServer.on('upgrade', (req, socket, head) => {
                this.websocketServer.handleUpgrade(req, socket, head, (ws) => {
                    this.websocketServer.emit('connection', ws, req);
                });
            });
            this.httpServer.on('listening', () => {
                this.log && this.log.info(this.httpServer.address(), `Statistics server started`);
            });
            this.httpServer.listen(this.port, this.host);
        };
        /**
         * Respond to the connections to websocket server
         */
        this.createWebsocketServer = () => {
            this.websocketServer = new ws_1.default.Server({ noServer: true });
            this.websocketServer.on('connection', (ws) => {
                let timer;
                const handleOpen = () => {
                    if (timer) {
                        clearInterval(timer);
                        timer = null;
                    }
                    if (this.updateInterval) {
                        timer = setInterval(processUpdate, this.updateInterval);
                    }
                };
                const handleClose = () => {
                    if (timer) {
                        clearInterval(timer);
                        timer = null;
                    }
                    ws.close();
                };
                /**
                 * Process timer expirations. Send out the set of statistics processed through the filters.
                 */
                const processUpdate = () => {
                    if (this.stats) {
                        const table = this.stats.getTable();
                        /**
                         * If there is a filter use it. Otherwise send everything
                         */
                        if (this.filter && Array.isArray(this.filter) && this.filter.length) {
                            const properties = Object.keys(table);
                            const result = {};
                            for (const property of properties) {
                                /**
                                 * Filtering is based on matching a filter to the start of the statistics name.
                                 * The workerId is skipped before the match is checked
                                 */
                                if (this.filter.find((filter) => property.substr(property.indexOf(':') + 1, filter.length) === filter)) {
                                    result[property] = table[property];
                                }
                            }
                            ws.send(JSON.stringify(result));
                        }
                        else {
                            ws.send(JSON.stringify(table));
                        }
                    }
                };
                ws.on('open', handleOpen);
                /**
                 * Handle messages from the websocket client.
                 * Client can start or stop the sending of data,
                 * change the update interval or set the filters
                 */
                ws.on('message', (message) => {
                    const data = JSON.parse(message);
                    if (data && data.command) {
                        switch (data.command) {
                            case 'start':
                                handleOpen();
                                processUpdate();
                                break;
                            case 'stop':
                                handleClose();
                                break;
                            case 'setInterval':
                                this.updateInterval = Number(data.interval);
                                break;
                            case 'setFilter':
                                this.filter = data.setFilter;
                                break;
                        }
                    }
                });
                ws.on('close', handleClose);
            });
        };
        if (options) {
            this.setOptions(options);
            if (!options.noStart) {
                this.start();
            }
        }
        else {
            throw Error('Missing configuration options for Statistics server');
        }
    }
}
exports.StatisticsServer = StatisticsServer;
//# sourceMappingURL=statisticsServer.js.map